@@ -27,7 +27,7 @@ Refactored heavily by Dan Sully
 #include "ape.h"
 
 static int _ape_error(ApeTag *tag, char *error, int ret) {
-  PerlIO_printf(PerlIO_stderr(), "APE: [%s] %s\n", error, tag->filename);
+  LOG_WARN("APE: [%s] %s\n", error, tag->filename);
   return ret;
 }
 
@@ -192,6 +192,9 @@ int _ape_get_tag_info(ApeTag* tag) {
   if (PerlIO_seek(tag->fd, (file_size -(long)tag->size - id3_length - (lyrics_size ? (lyrics_size + 15) : 0)), SEEK_SET) == -1) {
     return _ape_error(tag, "Couldn't seek to tag offset", -1);
   }
+  
+  tag->offset = file_size -(long)tag->size - id3_length - (lyrics_size ? (lyrics_size + 15) : 0);
+  DEBUG_TRACE("APE tag offset %d\n", tag->offset);
 
   /* ---------- Read tag header and data --------------- */
   buffer_init(&tag->tag_header, APE_TAG_HEADER_LEN);
@@ -226,12 +229,24 @@ int _ape_get_tag_info(ApeTag* tag) {
       return _ape_error(tag, "Couldn't seek to tag offset", -1);
     }
   }
+  
+  tag->offset += APE_TAG_HEADER_LEN;
 
   if (!_check_buf(tag->fd, &tag->tag_data, data_size, data_size)) {
     return _ape_error(tag, "Couldn't read tag data", -2);
   }
   
   tag->flags |= APE_CHECKED_APE | APE_HAS_APE;
+  
+  // Reduce the size of the audio_size value
+  if (my_hv_exists(tag->info, "audio_size")) {
+    int audio_size = SvIV(*(my_hv_fetch(tag->info, "audio_size")));
+    if (lyrics_size > 0)
+      lyrics_size += 15;
+    
+    my_hv_store(tag->info, "audio_size", newSVuv(audio_size - tag->size - lyrics_size));
+    DEBUG_TRACE("Reduced audio_size value by APE/Lyrics2 tag size %d\n", tag->size + lyrics_size);
+  }
 
   return 1;
 }
@@ -239,20 +254,14 @@ int _ape_get_tag_info(ApeTag* tag) {
 int _ape_parse_fields(ApeTag* tag) {
   int ret = 0;
   uint32_t i;
-  uint32_t offset = 0;
-  uint32_t last_possible_offset = tag->size - APE_MINIMUM_TAG_SIZE - APE_ITEM_MINIMUM_SIZE;
+
+  /* Don't exceed the maximum number of items allowed */
+  if (tag->num_fields >= APE_MAXIMUM_ITEM_COUNT) {
+    return _ape_error(tag, "Maximum item count exceeded", -3);
+  }
   
   for (i = 0; i < tag->item_count; i++) {
-    if (offset > last_possible_offset) {
-      return _ape_error(tag, "End of tag reached but more items specified", -3);
-    }
-
-    /* Don't exceed the maximum number of items allowed */
-    if (tag->num_fields == APE_MAXIMUM_ITEM_COUNT) {
-      return _ape_error(tag, "Maximum item count exceeded", -3);
-    }
-
-    if ((ret = _ape_parse_field(tag, &offset)) != 0) {
+    if ((ret = _ape_parse_field(tag)) != 0) {
       return ret;
     }
   }
@@ -266,7 +275,7 @@ int _ape_parse_fields(ApeTag* tag) {
   return 0;
 }
 
-int _ape_parse_field(ApeTag* tag, uint32_t* offset) {
+int _ape_parse_field(ApeTag* tag) {
 
   /* Ape tag item format:
    *
@@ -281,6 +290,9 @@ int _ape_parse_field(ApeTag* tag, uint32_t* offset) {
   SV *key = NULL;
   SV *value = NULL;
   
+  if (buffer_len(&tag->tag_data) < 8)
+    return _ape_error(tag, "Ran out of tag data before number of items was reached", -3);
+  
   size  = buffer_get_int_le(&tag->tag_data);
   flags = buffer_get_int_le(&tag->tag_data);
 
@@ -300,7 +312,9 @@ int _ape_parse_field(ApeTag* tag, uint32_t* offset) {
     tmp_ptr    += 1;
   }
   
-  DEBUG_TRACE("key_length: %d / val_length: %d / size: %d / flags %x\n", key_length, val_length, size, flags);
+  tag->offset += 8 + key_length + 1;
+  
+  DEBUG_TRACE("key_length: %d / val_length: %d / size: %d / flags %x @ %d\n", key_length, val_length, size, flags, tag->offset);
   
   if (flags & APE_TAG_TYPE_BINARY) {
     // Binary data, just copy it as-is
@@ -311,6 +325,9 @@ int _ape_parse_field(ApeTag* tag, uint32_t* offset) {
       if ( _env_true("AUDIO_SCAN_NO_ARTWORK") ) {
         // Don't read artwork, just return the size
         value = newSVuv(size - (val_length + 1) );
+        
+        my_hv_store( tag->tags, "COVER ART (FRONT)_offset", newSVuv(tag->offset + val_length + 1) );
+        
         buffer_consume(&tag->tag_data, size);
       }
       else {
@@ -323,6 +340,8 @@ int _ape_parse_field(ApeTag* tag, uint32_t* offset) {
       value = newSVpvn( buffer_ptr(&tag->tag_data), size );
       buffer_consume(&tag->tag_data, size);
     }
+    
+    tag->offset += val_length + 1;
   }
   else if (val_length >= size - 1) {
     // Single item
@@ -339,6 +358,8 @@ int _ape_parse_field(ApeTag* tag, uint32_t* offset) {
       sv_utf8_decode(value);
       DEBUG_TRACE("  %s = %s\n", SvPVX(key), SvPVX(value));
     }
+    
+    tag->offset += val_length < size ? val_length : size;
   }
   else {
     // Multiple items
@@ -357,6 +378,8 @@ int _ape_parse_field(ApeTag* tag, uint32_t* offset) {
       
       tmp_val = newSVpvn( buffer_ptr(&tag->tag_data), val_length );
       buffer_consume(&tag->tag_data, val_length);
+      
+      tag->offset += val_length;
     
       // Don't add invalid items
       if (_ape_check_validity(tag, flags, SvPVX(key), SvPVX(tmp_val)) != 0) {
@@ -375,6 +398,7 @@ int _ape_parse_field(ApeTag* tag, uint32_t* offset) {
       if ( done < size ) {
         // Still more to read, consume the null separator
         buffer_consume(&tag->tag_data, 1);
+        tag->offset++;
         done++;
       }
     }
@@ -475,6 +499,7 @@ get_ape_metadata(PerlIO *infile, char *file, HV *info, HV *tags)
   tag->filename   = file;
   tag->flags      = 0;
   tag->size       = 0;
+  tag->offset     = 0;
   tag->item_count = 0;
   tag->num_fields = 0;
 
